"""Test file foor pre_process.py module"""

import pytest
import numpy as np
import pandas as pd
import sklearn.pipeline
from lpsds.utils import keep, ObjectView, to_list, smart_tuple, confusion_matrix_annotation, pretty_title, error_bar_roc, pipeline_split, loc_multiindex

        
class TestKeep():
    @pytest.fixture
    def df(self):
        data = {}
        data['a'] = [11,22,33,44,55]
        data['b'] = [111,222,333,444,555]
        data['c'] = [True,True,False,True,False]
        data['d'] = ['v1','v2','v3','v4','v5']
        return pd.DataFrame(data)

    def test_remove_index(self, df):
        to_keep = df.a.isin([22,44])
        keep(df, index=to_keep)
        assert len(df) == 2
        assert df.index[0] == 1
        assert df.index[1] == 3

    def test_remove_column(self, df):
        to_keep = df.columns.isin(['a','d'])
        keep(df, columns=to_keep)
        assert len(df.columns) == 2
        assert df.columns[0] == 'a'
        assert df.columns[1] == 'd'
    
    def test_remove_index_and_columns(self, df):
        idx_to_keep = df.a.isin([11,22,55])
        cols_to_keep = df.columns == 'c'
        keep(df, index=idx_to_keep, columns=cols_to_keep)
        
        #Checking index
        assert len(df) == 3
        assert df.index[0] == 0
        assert df.index[1] == 1
        assert df.index[2] == 4
        
        #Checking columns
        assert len(df.columns) == 1
        assert df.columns[0] == 'c'

    def test_inplace_true(self, df):
        idx_to_keep = df.a.isin([33,44])
        ret = keep(df, index=idx_to_keep, inplace=True)
        assert ret is None

    def test_inplace_false(self, df):
        idx_to_keep = df.a.isin([33,44])
        ret = keep(df, index=idx_to_keep, inplace=False)
        
        #Ckecking that ret is not None and that df is unchanged.
        assert ret is not None
        assert len(df) == 5
        assert len(ret) == 2


class TestObjectView:
    """Test ObjectView class"""

    @pytest.fixture
    def obj(self):
        """Default ObjectView object"""
        return ObjectView()

    def test_set_value(self, obj):
        """Test if the obj.var =  value works"""
        obj.my_field = 10
        assert obj['my_field'] == 10

    def test_get_value(self, obj):
        """Test if we can read the value using the notation 'obj.var'"""
        obj['my_field'] = 10
        assert obj.my_field == 10

    def test_attribute_not_found(self, obj):
        """Test if an AttributeError is raised if a required field does not exists."""
        obj.my_field = 10
        with pytest.raises(AttributeError):
            obj.my_other_field
    
    def test_attribute_not_found_for_deletion(self, obj):
        """Test if an AttributeError is raised if a required field does not exists upon deletion."""
        obj.my_field = 10
        with pytest.raises(AttributeError):
            del obj.my_other_field



class TestToList:
    """Test to_list function"""

    def test_value_to_list(self):
        """Test if a value gets converted to a list"""
        ret = to_list(10)
        assert type(ret) is list
        assert len(ret) == 1
        assert ret[0] == 10

    def assert_list_remains_as_list(self):
        """A passed list must be returned unchanged"""
        val = [1]
        ret = to_list(val)
        assert type(ret) is list
        assert len(ret) == 1
        assert ret[0] == 10

    def assert_list_remains_as_tuple(self):
        """A passed tuple must be returned unchanged"""
        val = (10, 20)
        ret = to_list(val)
        assert type(ret) is tuple
        assert len(ret) == 2
        assert ret[0] == 10
        assert ret[1] == 20

    def assert_string_remains_as_string(self):
        """A passed string must be returned unchanged"""
        val = "my test string"
        ret = to_list(val)
        assert type(ret) is str
        assert ret == val


class TestSmartTuple:
    """Test smart_tuple function"""

    def test_single_value_tuple(self):
        val = (10,)
        assert smart_tuple(val) == 10

    def test_multi_value_tuple(self):
        val = (10, 20, 30)
        ret = smart_tuple(val)
        assert type(ret) is tuple
        assert len(ret) == 3
        assert ret[0] == 10
        assert ret[1] == 20
        assert ret[2] == 30



class TestCMAnnotate:

    @pytest.fixture
    def cm(self):
        return np.array([
            [[1,7,20],[40,5,16]],
            [[19,1,9], [10,21,2]],
            [[17,14,15],[26,17,18]],
            [[19,40,21], [32,23,24]]
        ])
    
    @pytest.fixture
    def fmt_str(self):
        return '${mean:.2f}^{{+{e_max:.2f}}}_{{-{e_min:.2f}}}$'
    
    @pytest.fixture
    def seed(self):
        return 23
    
    
    def test_ret_shape(self, cm):
        ret = confusion_matrix_annotation(cm)
        assert len(ret.shape) == 2
        assert ret.shape[0] == 2
        assert ret.shape[1] == 3
    

    def test_difference_false(self, cm, fmt_str, seed):
        ret = confusion_matrix_annotation(cm, use_difference=False, seed=seed)
        assert ret.iloc[0][0] == fmt_str.format(mean=14.00, e_min=5.49, e_max=19.00)
        assert ret.iloc[0][1] == fmt_str.format(mean=15.50, e_min=4.00, e_max=31.75)
        assert ret.iloc[0][2] == fmt_str.format(mean=16.25, e_min=11.75, e_max=20.50)

        assert ret.iloc[1][0] == fmt_str.format(mean=27.00, e_min=15.50, e_max=36.50)
        assert ret.iloc[1][1] == fmt_str.format(mean=16.50, e_min=8.98, e_max=22.00)
        assert ret.iloc[1][2] == fmt_str.format(mean=15.00, e_min=6.00, e_max=22.00)


    def test_difference_true(self, cm, fmt_str, seed):
        ret = confusion_matrix_annotation(cm, use_difference=True, seed=seed)
        assert ret.iloc[0][0] == fmt_str.format(mean=14.00, e_min=8.51, e_max=5.00)
        assert ret.iloc[0][1] == fmt_str.format(mean=15.50, e_min=11.50, e_max=16.25)
        assert ret.iloc[0][2] == fmt_str.format(mean=16.25, e_min=4.5, e_max=4.25)

        assert ret.iloc[1][0] == fmt_str.format(mean=27.00, e_min=11.50, e_max=9.50)
        assert ret.iloc[1][1] == fmt_str.format(mean=16.50, e_min=7.52, e_max=5.50)
        assert ret.iloc[1][2] == fmt_str.format(mean=15.00, e_min=9.00, e_max=7.00)


    def test_fmt_str(self, cm, seed):
        fmt_str = '{mean:.0f}, {e_min:.0f}, {e_max:.0f}'
        ret = confusion_matrix_annotation(cm, use_difference=True, seed=seed, fmt_str=fmt_str)
        assert ret.iloc[0][0] == fmt_str.format(mean=14, e_min=9, e_max=5)
        assert ret.iloc[0][1] == fmt_str.format(mean=16, e_min=12, e_max=16)
        assert ret.iloc[0][2] == fmt_str.format(mean=16, e_min=4, e_max=4)

        assert ret.iloc[1][0] == fmt_str.format(mean=27, e_min=12, e_max=10)
        assert ret.iloc[1][1] == fmt_str.format(mean=16, e_min=8, e_max=6)
        assert ret.iloc[1][2] == fmt_str.format(mean=15, e_min=9, e_max=7)


class TestPrettyTitle:

    @pytest.mark.parametrize(("val", "ret"), [ 
                                                ('auc', 'AUC'),
                                                ('Auc', 'AUC'),
                                                ('aUc', 'AUC'),
                                                ('auC', 'AUC'),
                                                ('AUc', 'AUC'),
                                                ('aUC', 'AUC'),
                                                ('AUC', 'AUC'),
                                                ('specificity', 'Specificity'),
                                                ('Specificity', 'Specificity'),
                                                ('SPECIFICITY', 'Specificity'),
                                            ])
    def test_normal_case(self, val, ret):
        assert pretty_title(val) == ret


    @pytest.mark.parametrize(("val", "ret"), [ 
                                                ('auc', 'Auc'),
                                                ('Auc', 'Auc'),
                                                ('aUc', 'Auc'),
                                                ('auC', 'Auc'),
                                                ('AUc', 'Auc'),
                                                ('aUC', 'Auc'),
                                                ('AUC', 'Auc'),
                                                ('specificity', 'Specificity'),
                                                ('Specificity', 'Specificity'),
                                                ('SPECIFICITY', 'Specificity'),
                                            ])
    def test_new_threshold(self, val, ret):
        assert pretty_title(val, abbrev_threshold=2) == ret


    @pytest.mark.parametrize(("val", "ret"), [ 
                                                ('auc', 'AUC'),
                                                ('Auc', 'AUC'),
                                                ('aUc', 'AUC'),
                                                ('auC', 'AUC'),
                                                ('AUc', 'AUC'),
                                                ('aUC', 'AUC'),
                                                ('AUC', 'AUC'),
                                                ('specificity', 'SPECIFICITY'),
                                                ('Specificity', 'SPECIFICITY'),
                                                ('SPECIFICITY', 'SPECIFICITY'),
                                            ])
    def test_new_threshold_large(self, val, ret):
        assert pretty_title(val, abbrev_threshold=20) == ret




class TestErrorBarRoc:

    @pytest.fixture
    def roc_df(self):
        df = pd.DataFrame(columns=['fold', 'false_positive', 'true_positive', 'threshold'])
        df.loc[len(df)] = 0, 0, 0, 0 # tp = 2*fp
        df.loc[len(df)] = 0, 0.2, 0.4, 0
        df.loc[len(df)] = 0, 0.3, 0.6, 0
        df.loc[len(df)] = 0, 1.0, 2.0, 0
        df.loc[len(df)] = 1, 0, 0, 0 # tp = 10*fp
        df.loc[len(df)] = 1, 0.6, 6, 0
        df.loc[len(df)] = 1, 1, 10, 0
        return df

    @pytest.fixture
    def ret(self, roc_df):
        return error_bar_roc(roc_df, num_points=5)

    def test_ret_size(self, ret):
        assert len(ret) == 10


    def test_col_names(self, ret):
        assert len(ret.columns) == 3
        assert ret.columns[0] == 'false_positive'
        assert ret.columns[1] == 'true_positive'
        assert ret.columns[2] == 'fold'
    
    def test_fold_0(self, ret):
        fold = ret.loc[ret.fold==0]
        assert len(fold) == 5
        assert fold.iloc[0].true_positive == 0
        assert fold.iloc[1].true_positive == 0.5
        assert fold.iloc[2].true_positive == 1
        assert fold.iloc[3].true_positive == 1.5
        assert fold.iloc[4].true_positive == 2


    def test_fold_1(self, ret):
        fold = ret.loc[ret.fold==1]
        assert len(fold) == 5
        assert fold.iloc[0].true_positive == 0
        assert fold.iloc[1].true_positive == 2.5
        assert fold.iloc[2].true_positive == 5.0
        assert fold.iloc[3].true_positive == 7.5
        assert fold.iloc[4].true_positive == 10

    def test_fp_(self, ret):
        fold0 = ret.loc[ret.fold==0]
        assert fold0.iloc[0].false_positive == 0
        assert fold0.iloc[1].false_positive == 0.25
        assert fold0.iloc[2].false_positive == 0.5
        assert fold0.iloc[3].false_positive == 0.75
        assert fold0.iloc[4].false_positive == 1.0


    def test_fp_equal(self, ret):
        fold0 = ret.loc[ret.fold==0]
        fold1 = ret.loc[ret.fold==1]
        assert fold0.iloc[0].false_positive == fold1.iloc[0].false_positive
        assert fold0.iloc[1].false_positive == fold1.iloc[1].false_positive
        assert fold0.iloc[2].false_positive == fold1.iloc[2].false_positive
        assert fold0.iloc[3].false_positive == fold1.iloc[3].false_positive
        assert fold0.iloc[4].false_positive == fold1.iloc[4].false_positive
    


class TestPipelineSplit:

    class PreProc1:
        def transform(self, X):
            return X + 1
    class PreProc2:
        def transform(self, X):
            return X + 10
    class Estimator:
        def predict(self, X):
            return np.array([11,22])


    @pytest.fixture
    def pipe(self):
        return sklearn.pipeline.Pipeline(steps = [
            ('pp1', TestPipelineSplit.PreProc1()),
            ('pp2', TestPipelineSplit.PreProc2()),
            ('est', TestPipelineSplit.Estimator()),
        ])
    
    def test_pre_proc_split(self, pipe):
        pp, mod, x = pipeline_split(pipe)
        assert len(pp) == 2
        assert pp[0][0] == 'pp1'
        assert isinstance(pp[0][1], TestPipelineSplit.PreProc1)
        assert pp[1][0] == 'pp2'
        assert isinstance(pp[1][1], TestPipelineSplit.PreProc2)


    def test_estimator(self, pipe):
        pp, mod, x = pipeline_split(pipe)
        assert mod[0] == 'est'
        assert isinstance(mod[1], TestPipelineSplit.Estimator)


    def test_x_none(self, pipe):
        pp, mod, x = pipeline_split(pipe)
        assert x is None


    def test_x(self, pipe):
        X = np.array([[1,2],
                      [3,4]])
        pp, mod, x = pipeline_split(pipe, X)
        assert X.shape[0] == 2
        assert X.shape[1] == 2
        assert x[0][0] == 12
        assert x[0][1] == 13
        assert x[1][0] == 14
        assert x[1][1] == 15


class TestLocMultiindex:
    """Test loc_multiindex function"""

    @pytest.fixture
    def simple_df(self):
        """Create a simple 2-level MultiIndex DataFrame"""
        df = pd.DataFrame({
            'a': [1, 1, 2, 2],
            'b': [4, 5, 4, 5],
            'value': [10, 20, 30, 40]
        })
        df.set_index(['a', 'b'], inplace=True)
        return df

    @pytest.fixture
    def three_level_df(self):
        """Create a 3-level MultiIndex DataFrame"""
        df = pd.DataFrame({
            'x': [1, 1, 1, 2, 2, 2],
            'y': ['A', 'A', 'B', 'A', 'A', 'B'],
            'z': [10, 20, 10, 10, 20, 10],
            'value': [100, 200, 300, 400, 500, 600]
        })
        df.set_index(['x', 'y', 'z'], inplace=True)
        return df

    def test_query_with_dict_found(self, simple_df):
        """Test querying with dictionary when index exists"""
        result = loc_multiindex(simple_df, {'a': 1, 'b': 4})
        assert isinstance(result, pd.Series)
        assert result['value'] == 10

    def test_query_with_dict_order_independent(self, simple_df):
        """Test that dictionary query order doesn't matter"""
        result1 = loc_multiindex(simple_df, {'a': 1, 'b': 5})
        result2 = loc_multiindex(simple_df, {'b': 5, 'a': 1})
        assert result1['value'] == result2['value']
        assert result1['value'] == 20

    def test_query_with_tuple_found(self, simple_df):
        """Test querying with tuple when index exists"""
        result = loc_multiindex(simple_df, (2, 4))
        assert isinstance(result, pd.Series)
        assert result['value'] == 30

    def test_query_with_tuple_order_matters(self, simple_df):
        """Test that tuple query order must match index order"""
        result = loc_multiindex(simple_df, (2, 5))
        assert result['value'] == 40

    def test_query_not_found_returns_default(self, simple_df):
        """Test that missing index returns default value (np.nan)"""
        result = loc_multiindex(simple_df, {'a': 99, 'b': 99})
        assert np.isnan(result)

    def test_query_not_found_custom_default(self, simple_df):
        """Test that missing index returns custom default value"""
        result = loc_multiindex(simple_df, {'a': 99, 'b': 99}, default='Not Found')
        assert result == 'Not Found'

    def test_query_not_found_custom_default_tuple(self, simple_df):
        """Test that missing tuple index returns custom default value"""
        result = loc_multiindex(simple_df, (99, 99), default=-1)
        assert result == -1

    def test_three_level_dict_query(self, three_level_df):
        """Test querying 3-level MultiIndex with dictionary"""
        result = loc_multiindex(three_level_df, {'x': 1, 'y': 'A', 'z': 10})
        assert isinstance(result, pd.Series)
        assert result['value'] == 100

    def test_three_level_tuple_query(self, three_level_df):
        """Test querying 3-level MultiIndex with tuple"""
        result = loc_multiindex(three_level_df, (2, 'B', 10))
        assert isinstance(result, pd.Series)
        assert result['value'] == 600

    def test_three_level_dict_order_independent(self, three_level_df):
        """Test that dictionary order doesn't matter for 3-level index"""
        result1 = loc_multiindex(three_level_df, {'x': 2, 'y': 'A', 'z': 20})
        result2 = loc_multiindex(three_level_df, {'z': 20, 'x': 2, 'y': 'A'})
        result3 = loc_multiindex(three_level_df, {'y': 'A', 'z': 20, 'x': 2})
        assert result1['value'] == result2['value'] == result3['value'] == 500

    def test_missing_index_level_raises_error(self, simple_df):
        """Test that missing index level in dict raises KeyError"""
        with pytest.raises(KeyError, match="Missing index levels"):
            loc_multiindex(simple_df, {'a': 1})

    def test_extra_keys_ignored(self, simple_df):
        """Test that extra keys in dict are ignored"""
        # Extra keys should be ignored, only index levels matter
        result = loc_multiindex(simple_df, {'a': 1, 'b': 4, 'c': 100, 'd': 200})
        assert isinstance(result, pd.Series)
        assert result['value'] == 10

    def test_empty_dataframe(self):
        """Test with empty MultiIndex DataFrame"""
        df = pd.DataFrame(columns=['value'])
        df.index = pd.MultiIndex.from_tuples([], names=['a', 'b'])
        result = loc_multiindex(df, {'a': 1, 'b': 2}, default='empty')
        assert result == 'empty'

    def test_single_level_index_with_dict(self):
        """Test that function works with single-level index using dict"""
        df = pd.DataFrame({'a': [1, 2, 3], 'value': [10, 20, 30]})
        df.set_index('a', inplace=True)
        result = loc_multiindex(df, {'a': 2})
        assert result['value'] == 20

    def test_single_level_index_with_tuple(self):
        """Test that function works with single-level index using tuple"""
        df = pd.DataFrame({'a': [1, 2, 3], 'value': [10, 20, 30]})
        df.set_index('a', inplace=True)
        result = loc_multiindex(df, (3,))
        assert result['value'] == 30

    def test_multiple_rows_returned(self):
        """Test behavior when query matches multiple rows"""
        # Create a DataFrame where one level can match multiple rows
        df = pd.DataFrame({
            'a': [1, 1, 2, 2],
            'b': [4, 5, 4, 5],
            'c': ['x', 'y', 'z', 'w'],
            'value': [10, 20, 30, 40]
        })
        df.set_index(['a', 'b'], inplace=True)
        
        # Query for a specific row
        result = loc_multiindex(df, {'a': 1, 'b': 4})
        assert isinstance(result, pd.Series)
        assert result['value'] == 10

    def test_numeric_string_mixed_index(self):
        """Test with mixed numeric and string index levels"""
        df = pd.DataFrame({
            'num': [1, 1, 2, 2],
            'str': ['alpha', 'beta', 'alpha', 'beta'],
            'value': [100, 200, 300, 400]
        })
        df.set_index(['num', 'str'], inplace=True)
        
        result = loc_multiindex(df, {'num': 2, 'str': 'alpha'})
        assert result['value'] == 300

    def test_default_none(self, simple_df):
        """Test with None as default value"""
        result = loc_multiindex(simple_df, {'a': 99, 'b': 99}, default=None)
        assert result is None

    def test_default_zero(self, simple_df):
        """Test with 0 as default value"""
        result = loc_multiindex(simple_df, {'a': 99, 'b': 99}, default=0)
        assert result == 0

    def test_default_dict(self, simple_df):
        """Test with dictionary as default value"""
        result = loc_multiindex(simple_df, {'a': 99, 'b': 99}, default={'error': 'not found'})
        assert result == {'error': 'not found'}

    def test_tuple_with_fewer_elements(self, simple_df):
        """Test tuple with fewer elements returns DataFrame (partial index match)"""
        # Too few elements - pandas returns DataFrame with all matching rows
        result = loc_multiindex(simple_df, (1,))
        assert isinstance(result, pd.DataFrame)
        assert len(result) == 2  # Both rows with a=1
        assert result.iloc[0]['value'] == 10
        assert result.iloc[1]['value'] == 20
        
    def test_tuple_with_more_elements_raises_error(self, simple_df):
        """Test tuple with too many elements raises IndexingError"""
        # Too many elements should raise an error (not caught by try-except)
        with pytest.raises(Exception):  # pandas.core.indexing.IndexingError
            loc_multiindex(simple_df, (1, 2, 3))

    def test_preserves_original_dataframe(self, simple_df):
        """Test that the function doesn't modify the original DataFrame"""
        original_len = len(simple_df)
        original_cols = list(simple_df.columns)
        
        loc_multiindex(simple_df, {'a': 1, 'b': 4})
        
        assert len(simple_df) == original_len
        assert list(simple_df.columns) == original_cols



import os
import datetime
import numpy as np
import pandas as pd
import pytest
import mlflow
import sklearn
import sklearn.utils
import lpsds.metrics
from lpsds.mlflow import MLFlow
from lpsds.utils import ObjectView

class MLFlowBase:

    class MockedRunClient:
        def get_metric_history(self, run_id, metric_name):

            sp_list = []
            f1_list = []
            for i in range(5):
                for j, v in enumerate([sp_list, f1_list]):
                    w = 1 if j==0 else 10
                    obj = ObjectView()
                    obj.value = w * (i+1)
                    obj.step = i+1
                    v.append(obj)

            ret_map = {}
            for m, v in zip(['sp_mean', 'f1_mean', 'acc'], [0.9, 0.95, 0.99]):
                obj = ObjectView()
                obj.value = v
                obj.step = 0
                ret_map[m] = [ obj ]

            ret_map['sp'] = sp_list
            ret_map['f1'] = f1_list
            return ret_map[metric_name]


    @pytest.fixture
    def mlf_obj(self):
        return MLFlow()

    @staticmethod
    def mocked_active_run():
        """
        Mocks mlflow.active_run behavior
        """
        ret = ObjectView()
        ret.info = ObjectView()
        ret.info.run_id = '112233'
        ret.info.experiment_id = '112233'
        return ret
    
    @staticmethod
    def mocked_get_run(run_id):
        """
        Mocks mlflow.active_run behavior
        """
        ret = ObjectView()
        ret.info = ObjectView()
        ret.info.run_id = run_id
        ret.info.experiment_id = run_id + '987'
        ret.data = ObjectView()
        ret.data.metrics = dict(
            sp_mean = 0.9,
            f1_mean = 0.95,
            acc = 0.99,
            sp = 0,
            f1 = 10,
        )
        ret.data.params = dict(
            param_str ='my string',
            param_int = '123',
            param_float = '3.14',
            param_bool = 'True',
            param_date = '2022-12-24',
        )
        return ret
    
    @staticmethod
    def mocked_get_experiment(exp_id):
        ret = ObjectView()
        ret.name = f'test_experiment_{exp_id}'
        return ret


    @staticmethod
    def mocked_download_artifacts(run_id, artifact_path, dst_path):
        df = pd.DataFrame({'a' : [1,2,3], 'b' : [11,22,33]})
        fname = os.path.split(artifact_path)[1]
        local_name = os.path.join(dst_path, fname)
        df.to_parquet(local_name)
        return local_name


    @pytest.fixture(autouse=True)
    def set_mlflow_patches(self, monkeypatch, request):
        if 'noautofixt' in request.keywords: return
        monkeypatch.setattr(mlflow, 'active_run', MLFlowBase.mocked_active_run)
        monkeypatch.setattr(mlflow, 'get_run', MLFlowBase.mocked_get_run)
        monkeypatch.setattr(mlflow, 'get_experiment', MLFlowBase.mocked_get_experiment)
        monkeypatch.setattr(mlflow.tracking, 'MlflowClient', MLFlowBase.MockedRunClient)
        monkeypatch.setattr(mlflow.artifacts, 'download_artifacts', MLFlowBase.mocked_download_artifacts)


class TestInit(MLFlowBase):

    def test_call_with_id(self):
        a = MLFlow('12345')
        assert a.run_id == '12345'

    @pytest.mark.noautofixt
    def test_call_without_id_without_experiment(self):
        with pytest.raises(ValueError):
            MLFlow()


    def test_call_without_id(self):
        a = MLFlow()
        assert a.run_id == '112233'


class TestLogStatistics(MLFlowBase):

    @staticmethod
    def mocked_log_metrics(metrics_map):
        """
        Mocks mlflow.log_metrics behavior
        """
        return None

    @staticmethod
    def mocked_bootstrap_estimate(vec, ci=95, n_boot=1000, seed=None):
        """
        Returns a mocked result for bootstrap_estimate
        """
        return vec.mean(), vec.mean()-vec.min(), vec.mean()+vec.max()
    
    @staticmethod
    def mocked_log_metric(key, value, step):
        return None

    @pytest.fixture(autouse=True)
    def set_patches(self, monkeypatch):
        monkeypatch.setattr(lpsds.metrics, 'bootstrap_estimate', TestLogStatistics.mocked_bootstrap_estimate)
        monkeypatch.setattr(mlflow, 'log_metrics', TestLogStatistics.mocked_log_metrics)
        monkeypatch.setattr(mlflow, 'log_metric', TestLogStatistics.mocked_log_metric)

    @pytest.fixture
    def cv_map(self):
        """
        Returns the mocked data for each test
        """        
        return {
            'test_sp' : np.array([1,2,3,4,5]), #mean = 3, err_min=2, err_max=8
            'estimators' : np.array([1,2,3,4,5]),
            'test_auc' : np.array([10,2,30,4,50]), #mean=19.2, err_min=17.2, err_max=69.2
        }

    def test_only_2_returned(self, cv_map, mlf_obj):
        """
        Only 2 metrics must be returned
        """
        ret = mlf_obj.log_statistics(cv_map)
        assert len(ret) == 2


    def test_metric_names(self, cv_map, mlf_obj):
        """
        Only SP and AUC must be returned
        """
        ret = mlf_obj.log_statistics(cv_map)
        assert 'sp' in ret
        assert 'auc' in ret

    def test_metric_values(self, cv_map, mlf_obj):
        """
        Only metric values must be correct
        """
        ret = mlf_obj.log_statistics(cv_map)
        assert ret['sp']['sp_mean'] == 3
        assert ret['sp']['sp_err_min'] == 2
        assert ret['sp']['sp_err_max'] == 8

        assert ret['auc']['auc_mean'] == 19.2
        assert ret['auc']['auc_err_min'] == 17.2
        assert ret['auc']['auc_err_max'] == 69.2
    
    @staticmethod
    def assert_metrics_logged(metric_map):
        assert metric_map['sp_mean'] == 3
        assert metric_map['sp_err_min'] == 2
        assert metric_map['sp_err_max'] == 8

    def test_mlflow_submission(self, monkeypatch, mlf_obj):
        met_map = {'test_sp' : np.array([1,2,3,4,5])} #mean = 3, err_min=2, err_max=8
        monkeypatch.setattr(mlflow, 'log_metrics', TestLogStatistics.assert_metrics_logged)
        mlf_obj.log_statistics(met_map)
    

    def test_metrics_vector_logged(self, monkeypatch, mlf_obj):
        self.metric_list = []
        def mocked_log_metric(key, value, step):
            self.metric_list.append((key, value, step))


        met_map = {'test_sp' : np.array([1,2,3,4,5])} #mean = 3, err_min=2, err_max=8
        monkeypatch.setattr(mlflow, 'log_metric', mocked_log_metric)
        mlf_obj.log_statistics(met_map)
        for i, items in enumerate(self.metric_list):
            k,v,s = items
            assert k == 'sp'
            assert v == i+1
            assert s == i



class TestLogArtifact(MLFlowBase):

    @staticmethod
    def assert_file_exists(temp_file_name, folder):
        assert os.path.exists(temp_file_name)
        assert folder == 'mlflow_test_folder'

    @staticmethod
    def assert_right_content(temp_file_name, folder):
        df = pd.read_parquet(temp_file_name)
        assert df.shape[0] == 3
        assert df.shape[1] == 2

        assert df.a.iloc[0] == 1
        assert df.a.iloc[1] == 2
        assert df.a.iloc[2] == 3

        assert df.b.iloc[0] == 40
        assert df.b.iloc[1] == 50
        assert df.b.iloc[2] == 60


    @pytest.fixture
    def df(self):
        return pd.DataFrame({'a' : [1,2,3], 'b' : [40,50,60]})

    def test_file_exists(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogArtifact.assert_file_exists)    
        mlf_obj.log_artifact(df, 'filename.parquet', 'mlflow_test_folder', save_func=df.to_parquet, fname_param_name='path', object_param_name=None)


    def test_file_correct(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogArtifact.assert_right_content)
        mlf_obj.log_artifact(df, 'filename.parquet', 'mlflow_test_folder', save_func=df.to_parquet, fname_param_name='path', object_param_name=None)



class TestGetRunID(MLFlowBase):
    def test_return(self, mlf_obj):
        assert mlf_obj.get_run_id() == '112233'


class TestGetMetrics(MLFlowBase):

    def test_num_returns(self, mlf_obj):
        met = mlf_obj.get_metrics(as_dict=True)
        assert len(met) == 5    

    def test_scalars(self, mlf_obj):
        met = mlf_obj.get_metrics(as_dict=True)
        assert met['sp_mean'] == 0.9
        assert met['f1_mean'] == 0.95
        assert met['acc'] == 0.99


    def test_object_view(self, mlf_obj):
        met = mlf_obj.get_metrics(as_dict=True)
        assert isinstance(met, ObjectView)


    def test_vectors(self, mlf_obj):
        met = mlf_obj.get_metrics(as_dict=True)
        assert np.array_equal(met['sp'], np.array([1,2,3,4,5]))
        assert np.array_equal(met['f1'], np.array([10,20,30,40,50]))


        
    def test_dataframe_structure(self, mlf_obj):
        met = mlf_obj.get_metrics()
        assert met.shape[0] == 13
        assert met.shape[1] == 3
        assert met.columns[0] == 'metric'
        assert met.columns[1] == 'step'
        assert met.columns[2] == 'value'

    

    def test_scalars_dataframe(self, mlf_obj):
        met = mlf_obj.get_metrics()
        assert met.loc[met.metric == 'sp_mean', 'value'].iloc[0] == 0.9
        assert met.loc[met.metric == 'f1_mean', 'value'].iloc[0] == 0.95
        assert met.loc[met.metric == 'acc', 'value'].iloc[0] == 0.99


    def test_vectors_dataframe(self, mlf_obj):
        met = mlf_obj.get_metrics()
        assert np.array_equal(met.loc[met.metric == 'sp', 'value'].to_numpy(), np.array([1,2,3,4,5]))
        assert np.array_equal(met.loc[met.metric == 'f1', 'value'].to_numpy(), np.array([10,20,30,40,50]))




class TestGetArtifact(MLFlowBase):

    def test_operation_no_folder(self, mlf_obj):
        df = mlf_obj.get_artifact('my_df', load_func=pd.read_parquet)
        assert df.shape[0] == 3
        assert df.shape[1] == 2
        assert df.iloc[0].a == 1
        assert df.iloc[0].b == 11
        assert df.iloc[1].a == 2
        assert df.iloc[1].b == 22
        assert df.iloc[2].a == 3
        assert df.iloc[2].b == 33

    def test_operation_with_folder(self, mlf_obj):
        df = mlf_obj.get_artifact('my_df', 'folder/path', load_func=pd.read_parquet)
        assert df.shape[0] == 3
        assert df.shape[1] == 2
        assert df.iloc[0].a == 1
        assert df.iloc[0].b == 11
        assert df.iloc[1].a == 2
        assert df.iloc[1].b == 22
        assert df.iloc[2].a == 3
        assert df.iloc[2].b == 33


class TestGetExperiment(MLFlowBase):
    def test_positive_case(self):
        a = MLFlow('123')
        assert a.get_experiment().name == 'test_experiment_123987'


class TestGetParams(MLFlowBase):

    def test_num_returns(self, mlf_obj):
        par = mlf_obj.get_params()
        assert len(par) == 5

    def test_str_return(self, mlf_obj):
        par = mlf_obj.get_params(infer_types=False)
        assert par['param_str'] == 'my string'
        assert par['param_int'] == '123'
        assert par['param_float'] == '3.14'
        assert par['param_bool'] == 'True'
        assert par['param_date'] == '2022-12-24'


    def test_infer_type_return(self, mlf_obj):
        par = mlf_obj.get_params(infer_types=True)
        assert par['param_str'] == 'my string'
        assert par['param_int'] == 123
        assert par['param_float'] == 3.14
        assert par['param_bool'] == True
        assert par['param_date'] == datetime.date(2022, 12, 24)



    def test_object_view(self, mlf_obj):
        par = mlf_obj.get_params()
        assert isinstance(par, ObjectView)


class TestLogDataFrame(MLFlowBase):
    @pytest.fixture
    def df(self):
        return pd.DataFrame({'a' : [1,2,3], 'b' : [40,50,60]})

    def test_file_exists(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogArtifact.assert_file_exists)    
        mlf_obj.log_dataframe(df, 'filename', 'mlflow_test_folder')


    def test_file_correct(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogArtifact.assert_right_content)
        mlf_obj.log_dataframe(df, 'filename', 'mlflow_test_folder')


class TestLogNumpy(MLFlowBase):
    @staticmethod
    def assert_file_exists(temp_file_name, folder):
        assert os.path.exists(temp_file_name)
        assert folder == 'mlflow_test_folder'

    @staticmethod
    def assert_right_content(temp_file_name, folder):
        df = np.load(temp_file_name)
        assert df.shape[0] == 2
        assert df.shape[1] == 3

        assert df[0][0] == 1
        assert df[0][1] == 2
        assert df[0][2] == 3

        assert df[1][0] == 40
        assert df[1][1] == 50
        assert df[1][2] == 60


    @pytest.fixture
    def df(self):
        return np.array([[1,2,3],[40,50,60]])

    def test_file_exists(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogNumpy.assert_file_exists)    
        mlf_obj.log_numpy(df, 'filename', 'mlflow_test_folder')


    def test_file_correct(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogNumpy.assert_right_content)
        mlf_obj.log_numpy(df, 'filename', 'mlflow_test_folder')



class TestGetDataFrame(MLFlowBase):

    def test_operation_no_folder(self, mlf_obj):
        df = mlf_obj.get_dataframe('my_df')
        assert df.shape[0] == 3
        assert df.shape[1] == 2
        assert df.iloc[0].a == 1
        assert df.iloc[0].b == 11
        assert df.iloc[1].a == 2
        assert df.iloc[1].b == 22
        assert df.iloc[2].a == 3
        assert df.iloc[2].b == 33

    def test_operation_with_folder(self, mlf_obj):
        df = mlf_obj.get_dataframe('my_df', 'folder/path')
        assert df.shape[0] == 3
        assert df.shape[1] == 2
        assert df.iloc[0].a == 1
        assert df.iloc[0].b == 11
        assert df.iloc[1].a == 2
        assert df.iloc[1].b == 22
        assert df.iloc[2].a == 3
        assert df.iloc[2].b == 33




class TestGetNumpy(MLFlowBase):

    @staticmethod
    def mocked_download_artifacts_numpy(run_id, artifact_path, dst_path):
        df = np.array([[1,2,3],[11,22,33]])
        fname = os.path.split(artifact_path)[1]
        local_name = os.path.join(dst_path, fname + '.npy')
        np.save(local_name, df)
        return local_name


    def test_operation_no_folder(self, mlf_obj, monkeypatch):
        monkeypatch.setattr(mlflow.artifacts, 'download_artifacts', TestGetNumpy.mocked_download_artifacts_numpy)
        df = mlf_obj.get_numpy('my_df')
        assert df.shape[0] == 2
        assert df.shape[1] == 3
        assert df[0][0] == 1
        assert df[0][1] == 2
        assert df[0][2] == 3
        assert df[1][0] == 11
        assert df[1][1] == 22
        assert df[1][2] == 33

    def test_operation_with_folder(self, mlf_obj, monkeypatch):
        monkeypatch.setattr(mlflow.artifacts, 'download_artifacts', TestGetNumpy.mocked_download_artifacts_numpy)
        df = mlf_obj.get_numpy('my_df', 'folder/path')
        assert df.shape[0] == 2
        assert df.shape[1] == 3
        assert df[0][0] == 1
        assert df[0][1] == 2
        assert df[0][2] == 3
        assert df[1][0] == 11
        assert df[1][1] == 22
        assert df[1][2] == 33


class TestLogPipeline(MLFlowBase):
    """Test class for log_pipeline method"""

    @staticmethod
    def mocked_set_config(display):
        """Mock sklearn.set_config to avoid side effects"""
        pass

    @staticmethod
    def mocked_estimator_html_repr(pipeline):
        """Mock HTML representation of pipeline"""
        return '<div class="sklearn-pipeline">Mock Pipeline HTML</div>'

    @staticmethod
    def mocked_log_artifact(file_path, artifact_path):
        """Mock mlflow.log_artifact and verify file contents"""
        # Verify file exists and has correct content
        assert os.path.exists(file_path)
        with open(file_path, 'r', encoding='utf-8') as f:
            content = f.read()
        
        # Basic HTML structure checks
        assert '<!DOCTYPE html>' in content
        assert '<title>SOM Pipeline' in content
        # Check for sklearn pipeline content (real HTML representation)
        assert 'Pipeline' in content
        return None

    @pytest.fixture
    def mock_pipeline(self):
        """Create a mock pipeline for testing"""
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler
        from sklearn.linear_model import LogisticRegression
        
        return Pipeline([
            ('scaler', StandardScaler()),
            ('classifier', LogisticRegression())
        ])

    def test_basic_pipeline_logging(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test basic pipeline logging with minimal parameters"""
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogPipeline.mocked_log_artifact)
        
        # Should not raise any exceptions
        mlf_obj.log_pipeline(mock_pipeline, 'Test Pipeline')

    def test_pipeline_with_folder(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test pipeline logging with custom folder path"""
        def assert_folder_path(file_path, artifact_path):
            assert artifact_path == 'custom/folder'
            TestLogPipeline.mocked_log_artifact(file_path, artifact_path)
        
        monkeypatch.setattr(mlflow, 'log_artifact', assert_folder_path)
        mlf_obj.log_pipeline(mock_pipeline, 'Test Pipeline', folder='custom/folder')

    def test_pipeline_with_metadata(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test pipeline logging with metadata dictionary"""
        def assert_metadata_in_html(file_path, artifact_path):
            with open(file_path, 'r', encoding='utf-8') as f:
                content = f.read()
            
            # Check that metadata is included in HTML
            assert '<strong>Author:</strong> Test User' in content
            assert '<strong>Version:</strong> 1.0' in content
            assert '<strong>Model Type:</strong> Classification' in content
            
        monkeypatch.setattr(mlflow, 'log_artifact', assert_metadata_in_html)
        
        metadata = {
            'Author': 'Test User',
            'Version': '1.0',
            'Model Type': 'Classification'
        }
        mlf_obj.log_pipeline(mock_pipeline, 'Test Pipeline', metadata=metadata)

    def test_pipeline_with_empty_metadata(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test pipeline logging with empty metadata dictionary"""
        def assert_no_metadata_section(file_path, artifact_path):
            with open(file_path, 'r', encoding='utf-8') as f:
                content = f.read()
            
            # Should still have run ID but no additional metadata
            assert 'Run ID:' in content
            
        monkeypatch.setattr(mlflow, 'log_artifact', assert_no_metadata_section)
        mlf_obj.log_pipeline(mock_pipeline, 'Test Pipeline', metadata={})

    def test_pipeline_with_none_metadata(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test pipeline logging with None metadata"""
        def assert_none_metadata(file_path, artifact_path):
            with open(file_path, 'r', encoding='utf-8') as f:
                content = f.read()
            
            # Should only have run ID, no additional metadata HTML
            assert 'Run ID:' in content
            # Should not have extra metadata paragraphs
            content_lines = content.split('\n')
            metadata_section = [line for line in content_lines if '<strong>' in line and 'Run ID' not in line]
            assert len(metadata_section) == 0
            
        monkeypatch.setattr(mlflow, 'log_artifact', assert_none_metadata)
        mlf_obj.log_pipeline(mock_pipeline, 'Test Pipeline', metadata=None)

    def test_html_file_structure(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test that generated HTML file has correct structure"""
        def assert_html_structure(file_path, artifact_path):
            with open(file_path, 'r', encoding='utf-8') as f:
                content = f.read()
            
            # Check HTML structure
            assert '<!DOCTYPE html>' in content
            assert '<html lang="en">' in content
            assert '<meta charset="UTF-8">' in content
            assert '<title>SOM Pipeline - 112233' in content  # Full run_id (6 chars)
            assert '<h1>🔧 Test Pipeline</h1>' in content
            assert '<h2>Pipeline Structure</h2>' in content
            assert 'Pipeline' in content  # Check for pipeline content
            
            # Check CSS styling
            assert 'background-color: #1e1e1e' in content
            assert 'color: #d4d4d4' in content
            assert 'font-family: monospace' in content
            
        monkeypatch.setattr(mlflow, 'log_artifact', assert_html_structure)
        mlf_obj.log_pipeline(mock_pipeline, 'Test Pipeline')

    def test_sklearn_config_called(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test that sklearn.set_config is called with display='diagram'"""
        config_calls = []
        
        def track_set_config(display):
            config_calls.append(display)
        
        # Patch in the lpsds.mlflow module where it's imported
        monkeypatch.setattr('lpsds.mlflow.sklearn.set_config', track_set_config)
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogPipeline.mocked_log_artifact)
        
        mlf_obj.log_pipeline(mock_pipeline, 'Test Pipeline')
        
        assert len(config_calls) == 1
        assert config_calls[0] == 'diagram'

    def test_estimator_html_repr_called(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test that estimator_html_repr is called with the pipeline"""
        html_repr_calls = []
        
        def track_html_repr(pipeline):
            html_repr_calls.append(pipeline)
            return '<div>Mock HTML</div>'
        
        # Patch in the lpsds.mlflow module where it's imported
        monkeypatch.setattr('lpsds.mlflow.estimator_html_repr', track_html_repr)
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogPipeline.mocked_log_artifact)
        
        mlf_obj.log_pipeline(mock_pipeline, 'Test Pipeline')
        
        assert len(html_repr_calls) == 1
        assert html_repr_calls[0] is mock_pipeline

    def test_file_name_consistency(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test that the HTML file is always named 'model_pipeline.html'"""
        def assert_filename(file_path, artifact_path):
            filename = os.path.basename(file_path)
            assert filename == 'model_pipeline.html'
        
        monkeypatch.setattr(mlflow, 'log_artifact', assert_filename)
        mlf_obj.log_pipeline(mock_pipeline, 'Test Pipeline')

    def test_special_characters_in_title(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test pipeline logging with special characters in title"""
        def assert_special_chars_handled(file_path, artifact_path):
            with open(file_path, 'r', encoding='utf-8') as f:
                content = f.read()
            
            # Check that special characters are handled (not escaped in f-string)
            assert '<h1>🔧 Test <Pipeline> & Model</h1>' in content
            
        monkeypatch.setattr(mlflow, 'log_artifact', assert_special_chars_handled)
        mlf_obj.log_pipeline(mock_pipeline, 'Test <Pipeline> & Model')

    def test_run_id_in_title(self, monkeypatch, mock_pipeline, mlf_obj):
        """Test that run ID is correctly truncated to 8 characters in title"""
        def assert_run_id_truncated(file_path, artifact_path):
            with open(file_path, 'r', encoding='utf-8') as f:
                content = f.read()
            
            # Run ID should be truncated to first 8 characters (or the full ID if shorter)
            assert 'SOM Pipeline - 112233' in content  # Full run_id (6 chars, less than 8)
            
        monkeypatch.setattr(mlflow, 'log_artifact', assert_run_id_truncated)
        mlf_obj.log_pipeline(mock_pipeline, 'Test Pipeline')



import os
import re
import sklearn
import yaml
import numpy as np
import pandas as pd
import tempfile
import mlflow
import lpsds.metrics
from typing import Dict, Optional, Any, Callable, Union
from lpsds.utils import ObjectView
from sklearn.utils import estimator_html_repr
from sklearn.pipeline import Pipeline


class MLFlow:
    """MLFlow utilities class for logging and retrieving artifacts, metrics, and models."""
    
    def __init__(self, run_id: Optional[str] = None) -> None:
        """Initialize MLFlow instance.
        
        Args:
            run_id: MLFlow run ID. If None, uses the currently active run.
            
        Raises:
            ValueError: If no run_id provided and no active run exists.
        """
        active_run = mlflow.active_run()

        if run_id is None and active_run is None:
            raise ValueError('You must invoke this class either by passing a valid run id or within an started run (mlflow.start_run)')

        self.run_id = active_run.info.run_id if run_id is None else run_id
        self.run = mlflow.get_run(self.run_id)
        self.run_client = mlflow.tracking.MlflowClient()

    def log_dataframe(self, df: pd.DataFrame, var_name: str, folder: str = '') -> None:
        """Save a pandas DataFrame to MLFlow as a parquet file.

        Args:
            df: The pandas DataFrame object to be saved.
            var_name: The name the dataframe will have within MLFlow.
            folder: The path (in MLFlow) to where the dataframe will be saved.
        """
        self.log_artifact(df, var_name + '.parquet', folder, save_func=df.to_parquet, object_param_name=None, fname_param_name='path')

    def log_numpy(self, mat: np.ndarray, var_name: str, folder: str = '') -> None:
        """Save a numpy array to MLFlow as a .npy file.

        Args:
            mat: The numpy array object to be saved.
            var_name: The name the array will have within MLFlow.
            folder: The path (in MLFlow) to where the array will be saved.
        """
        self.log_artifact(mat, var_name + '.npy', folder, allow_pickle=False)

    def log_artifact(self, obj: Any, var_name: str, folder: str = '', save_func: Callable = np.save,
                     object_param_name: Optional[str] = 'arr', fname_param_name: str = 'file', 
                     **save_func_kwargs: Any) -> None:
        """Save an object of any type to MLFlow.

        Args:
            obj: The object you want to save to MLFlow.
            var_name: The name the object will have within MLFlow.
            folder: The path (in MLFlow) to where the object will be saved.
            save_func: A function responsible for saving the object to disk.
            object_param_name: The save_func parameter name for the object. If None, 
                             obj is not passed as a named parameter.
            fname_param_name: The save_func parameter name for the file path.
            **save_func_kwargs: Additional parameters to be passed to save_func.
        """
        with tempfile.TemporaryDirectory() as temp_path:
            temp_file_name = os.path.join(temp_path, var_name)
            save_func_kwargs[fname_param_name] = temp_file_name
            if object_param_name is not None:
                save_func_kwargs[object_param_name] = obj
            save_func(**save_func_kwargs)
            mlflow.log_artifact(temp_file_name, folder)

    def log_statistics(self, cv_model: Dict[str, Any]) -> Dict[str, Dict[str, float]]:
        """Log cross-validation metrics to MLFlow as statistical summaries.
        
        Automatically finds metrics with the pattern "test_*" and logs their mean,
        minimum, and maximum values with 95% confidence intervals.

        Args:
            cv_model: The result of the sklearn cross_validate function.
        
        Returns:
            Dictionary mapping metric names to their statistical summaries 
            (mean, err_min, err_max).
        """
        regexp = re.compile(r'test_(.+)')
        ret_map = {}
        for metric, values in cv_model.items():
            grp = regexp.match(metric)
            if grp is not None:
                mean, err_min, err_max = lpsds.metrics.bootstrap_estimate(values)
                metric_name = grp.group(1)
                ret_map[metric_name] = {
                    f'{metric_name}_err_min': err_min,
                    f'{metric_name}_mean': mean,
                    f'{metric_name}_err_max': err_max
                }
                mlflow.log_metrics(ret_map[metric_name])
                for i, v in enumerate(values): 
                    mlflow.log_metric(metric_name, v, step=i)
        
        return ret_map

    def get_params(self, infer_types: bool = False) -> ObjectView:
        """Return model parameters.

        Args:
            infer_types: If True, will try to infer value types since MLFlow 
                        stores them as strings.
        
        Returns:
            ObjectView containing the model parameters.
        """
        ret = ObjectView(self.run.data.params)
        if infer_types:
            for k, v in ret.items():
                ret[k] = yaml.safe_load(v)
        return ret

    def get_run_id(self) -> str:
        """Return the experiment's run ID.
        
        Returns:
            The MLFlow run ID.
        """
        return self.run_id

    def get_metrics(self, as_dict: bool = False) -> Union[pd.DataFrame, ObjectView]:
        """Collect all metrics available for a given run.

        Args:
            as_dict: If True, return metrics as a dict where vector metrics 
                    are returned as numpy arrays. If False, return as DataFrame.

        Returns:
            Either a pandas DataFrame with all metrics or an ObjectView dict 
            with metrics as numpy arrays for vector metrics.
        """
        # Collecting all metrics in a dataframe
        df = pd.DataFrame(columns=['metric', 'step', 'value'])
        for metric_name in self.run.data.metrics.keys():
            for m in self.run_client.get_metric_history(self.run_id, metric_name):
                df.loc[len(df)] = metric_name, m.step, m.value
        df.sort_values(['metric', 'step'], inplace=True, ignore_index=True)
        
        if not as_dict: 
            return df
    
        # Creating a dataframe where vectorized metrics are saved as lists
        grp = df.groupby('metric').value.agg(lambda x: x.iloc[0] if len(x) == 1 else x.to_list()).to_dict()
        
        # Lists are converted to np.arrays
        for k, v in grp.items():
            if hasattr(v, '__iter__'):
                grp[k] = np.array(v)
        return ObjectView(grp)

    def get_dataframe(self, var_name: str, folder: str = '') -> pd.DataFrame:
        """Load a pandas DataFrame saved to MLFlow as a parquet file.

        Args:
            var_name: The name the dataframe has within MLFlow.
            folder: The path (in MLFlow) from where the dataframe will be loaded.
        
        Returns:
            pandas DataFrame with the loaded data.
        """
        return self.get_artifact(var_name + '.parquet', folder, pd.read_parquet)

    def get_numpy(self, var_name: str, folder: str = '') -> np.ndarray:
        """Load a numpy array saved to MLFlow as a .npy file.

        Args:
            var_name: The name the array has within MLFlow.
            folder: The path (in MLFlow) from where the array will be loaded.
        
        Returns:
            numpy ndarray with the loaded data.
        """
        return self.get_artifact(var_name + '.npy', folder, np.load, allow_pickle=False)

    def get_artifact(self, var_name: str, folder: str = '', load_func: Callable = np.load, 
                     **load_func_kwargs: Any) -> Any:
        """Load an artifact using the provided loading function.

        Args:
            var_name: The name the artifact has within MLFlow (must include extensions).
            folder: The path (in MLFlow) from where the artifact will be loaded.
            load_func: The function used to load the required artifact.
            **load_func_kwargs: Additional parameters to be passed to load_func.
        
        Returns:
            Whatever the load_func returns.
        """
        with tempfile.TemporaryDirectory() as temp_path:
            full_path = os.path.join(folder, var_name)
            local_path = mlflow.artifacts.download_artifacts(run_id=self.run_id, artifact_path=full_path, dst_path=temp_path)
            return load_func(local_path, **load_func_kwargs)

    def get_experiment(self) -> mlflow.entities.Experiment:
        """Return the experiment that contains the MLFlow run.
        
        Returns:
            MLFlow Experiment instance containing this run.
        """
        exp_id = self.run.info.experiment_id
        return mlflow.get_experiment(exp_id)

    def log_pipeline(self, pipeline: Pipeline, title: str, name: str = 'model_pipeline', folder: str = '',
                     metadata: Optional[Dict[str, str]] = None) -> None:
        """Log a scikit-learn pipeline as an HTML visualization.

        Args:
            pipeline: The scikit-learn Pipeline to visualize.
            title: Title for the HTML page.
            folder: The path (in MLFlow) where the HTML file will be saved.
            metadata: Optional dictionary of metadata to include in the HTML.
        """
        # Enable diagram display
        sklearn.set_config(display='diagram')

        # Generate HTML representation of pipeline
        pipeline_html = estimator_html_repr(pipeline)

        # Save to temporary file with specific name
        with tempfile.TemporaryDirectory() as temp_dir:
            pipeline_viz_file = os.path.join(temp_dir, f'{name}.html')

            meta_html = '\n'.join([f'<p><strong>{key}:</strong> {value}</p>' for key, value in metadata.items()]) if metadata else ''

            with open(pipeline_viz_file, 'w', encoding='utf-8') as f:
                f.write(f"""<!DOCTYPE html>
                    <html lang="en">
                    <head>
                        <meta charset="UTF-8">
                        <meta name="viewport" content="width=device-width, initial-scale=1.0">
                        <title>SOM Pipeline - {self.run.info.run_id[:8]}</title>
                        <style>
                            body {{ 
                                background-color: #1e1e1e; 
                                color: #d4d4d4;
                                padding: 20px;
                                font-family: monospace;
                            }}
                            h1 {{ color: #4fc3f7; }}
                            .metadata {{ 
                                background: #2d2d2d; 
                                padding: 15px; 
                                border-radius: 8px;
                                margin: 20px 0;
                            }}
                        </style>
                    </head>
                    <body>
                        <h1>🔧 {title}</h1>
                        <div class="metadata">
                            <p><strong>Run ID:</strong> {self.run.info.run_id}</p>
                            {meta_html}
                        </div>
                        <h2>Pipeline Structure</h2>
                        {pipeline_html}
                    </body>
                    </html>
                """)

            mlflow.log_artifact(pipeline_viz_file, artifact_path=folder)
